from fastapi import FastAPI, Request, Query
from fastapi.middleware.cors import CORSMiddleware
from sqlalchemy import create_engine
from pydantic import BaseModel, Field
from typing import List, Optional
from datetime import date
import uvicorn
import jwt


app = FastAPI(title='Profile Service API')
db = create_engine('postgresql://postgres:postgres@profile_service_db:5432/postgres')

app.add_middleware(
    CORSMiddleware,
    allow_origins=['http://localhost:3000'],
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)

PROFILES_URL = '/api/profiles'
PROFILE_URL = '/api/profile'
JWT_SECRET = 'auth_service_secret'
JWT_ALGORITHM = 'HS256'

class Profile(BaseModel):
    id: int = Field(description='ID')
    first_name: str = Field(description='First Name')
    last_name: str = Field(description='Last Name')
    email: str = Field(description='Email')
    phone_number: str = Field(description='Phone Number')
    sex: str = Field(description='Sex')
    birth_date: date = Field(description='Birth Date')
    username: str = Field(description='Username')
    biography: str = Field(description='Biography')
    private: bool = Field(description='Flag if profile is private')
    work_experiences: Optional[list] = Field(description='Work Experiences')
    educations: Optional[list] = Field(description='Educations')
    skills: Optional[list] = Field(description='Skills')
    interests: Optional[list] = Field(description='Interests')
    connections: Optional[list] = Field(description='Connections IDs')
    connection_requests: Optional[list] = Field(description='Connection Requests IDs')
    blocked_profiles: Optional[list] = Field(description='Blocked Profiles IDs')


class NavigationLinks(BaseModel):
    base: str = Field('http://localhost:8001/api', description='API base URL')
    prev: Optional[str] = Field(None, description='Link to the previous page')
    next: Optional[str] = Field(None, description='Link to the next page')

class Response(BaseModel):
    results: List[Profile]
    links: NavigationLinks
    offset: int
    limit: int
    size: int


def get_current_user(request: Request):
    try:
        return get_user(jwt.decode(request.headers['Authorization'], JWT_SECRET, algorithms=[JWT_ALGORITHM])['id'])
    except:
        return None

def get_user(id: int):
    with db.connect() as connection:
        return Profile.parse_obj(dict(list(connection.execute('select * from profiles where id=%s', (id,)))[0]))

def search_query(search: str):
    return ' '.join("""
        lower(first_name) like %s or 
        lower(last_name) like %s or 
        lower(email) like %s or 
        lower(phone_number) like %s or 
        lower(sex) like %s or 
        lower(username) like %s or 
        lower(biography) like %s
    """.split()), (f'%{search.lower()}%', f'%{search.lower()}%', f'%{search.lower()}%', f'%{search.lower()}%', f'%{search.lower()}%', f'%{search.lower()}%', f'%{search.lower()}%')

def search_profiles(request: Request, search: str, offset: int, limit: int, type: str = ''):
    current_user = get_current_user(request)
    current_user_id = current_user.id if current_user else -1
    query, params = search_query(search)
    if type == '':
        filter = 'true'
    elif type == '/public':
        filter = 'private=false'
    else:
        filter_list = current_user.connections if type == '/connections' else current_user.connection_requests
        filter = f"id in {str(filter_list).replace('[', '(').replace(']', ')')}" if filter_list else 'false'

    with db.connect() as connection:
        total_profiles = len(list(connection.execute(f'select * from profiles where id!={current_user_id} and {filter} and ({query})', params)))
    prev_link = f'/profiles{type}?search={search}&offset={offset - limit}&limit={limit}' if offset - limit >= 0 else None
    next_link = f'/profiles{type}?search={search}&offset={offset + limit}&limit={limit}' if offset + limit < total_profiles else None
    links = NavigationLinks(prev=prev_link, next=next_link)

    with db.connect() as connection:
        profiles = list(connection.execute(f'select * from profiles where id!={current_user_id} and {filter} and ({query}) order by id offset {offset} limit {limit}', params))
    results = [Profile.parse_obj(dict(profile)) for profile in profiles]
    return Response(results=results, links=links, offset=offset, limit=limit, size=len(results))



@app.get(PROFILES_URL)
def read_profiles(request: Request, search: str = Query(''), offset: int = Query(0), limit: int = Query(7)):
    return search_profiles(request, search, offset, limit)


@app.get(f"{PROFILES_URL}/public")
def read_public_profiles(request: Request, search: str = Query(''), offset: int = Query(0), limit: int = Query(7)):
    return search_profiles(request, search, offset, limit, "/public")


@app.get(f"{PROFILES_URL}/connections")
def read_connections(request: Request, search: str = Query(''), offset: int = Query(0), limit: int = Query(7)):
    return search_profiles(request, search, offset, limit, "/connections")


@app.get(f"{PROFILES_URL}/connection_requests")
def read_connection_requests(request: Request, search: str = Query(''), offset: int = Query(0), limit: int = Query(7)):
    return search_profiles(request, search, offset, limit, "/connection_requests")



def run_service():
    uvicorn.run(app, host="0.0.0.0", port=8001)


if __name__ == '__main__':
    run_service()
